class InvalidType(TypeError):
    """Raises when an invalid data type is passed
    """
    pass

class InvalidWindow(ValueError):
    """Raises when an invalid functioning window is passed
    """
    pass

